"""Define query strings used by the Ridwell API."""
QUERY_ACCOUNT_DATA = """
query user($id: ID!) {
  user(id: $id) {
    fullName
    email
    phone
    accounts {
      id
      address {
        street1
        city
        subdivision
        postalCode
      }
      activeSubscription {
        id
        state
      }
    }
  }
}
"""

QUERY_AUTH_DATA = """
mutation createAuthentication($input: CreateAuthenticationInput!) {
  createAuthentication(input: $input) {
    authenticationToken
  }
}
"""

QUERY_SUBSCRIPTION_PICKUP_QUOTE = """
query subscriptionPickupQuote($input: SubscriptionPickupQuoteInput!) {
  subscriptionPickupQuote(input: $input) {
    totalCents
  }
}
"""

QUERY_SUBSCRIPTION_PICKUPS = """
query upcomingSubscriptionPickups($subscriptionId: ID!) {
  upcomingSubscriptionPickups(subscriptionId: $subscriptionId) {
    id
    state
    pickupOn
    pickupProductSelections {
      pickupOfferPickupProduct {
        pickupOffer {
          id
          priority
          category {
            name
          }
        }
        pickupProduct {
          id
        }
      }
      quantity
    }
  }
}
"""



"""Define an API client."""
from __future__ import annotations

import asyncio
from dataclasses import dataclass, field
from datetime import date, datetime
from typing import Any, Callable, Dict, Literal, TypedDict, cast

from aiohttp import ClientSession, ClientTimeout
from aiohttp.client_exceptions import ClientError, ContentTypeError
import jwt
from titlecase import titlecase

from .const import LOGGER
from .errors import (
    InvalidCredentialsError,
    RequestError,
    RidwellError,
    TokenExpiredError,
    raise_for_data_error,
)
from .query import (
    QUERY_ACCOUNT_DATA,
    QUERY_AUTH_DATA,
    QUERY_SUBSCRIPTION_PICKUP_QUOTE,
    QUERY_SUBSCRIPTION_PICKUPS,
)

API_BASE_URL = "https://api.ridwell.com"

DEFAULT_RETRIES = 3
DEFAULT_RETRY_DELAY = 1
DEFAULT_TIMEOUT = 10

CATEGORY_ADD_ON = "add_on"
CATEGORY_ROTATING = "rotating"
CATEGORY_STANDARD = "standard"

EVENT_STATE_INITIALIIZED = "initialized"
EVENT_STATE_SCHEDULED = "scheduled"

PICKUP_TYPES_ADD_ON = [
    "Beyond the Bin",
    "Fluorescent Light Tubes",
    "Latex Paint",
    "Paint",
    "Styrofoam",
]
PICKUP_TYPES_STANDARD = [
    "Batteries",
    "Light Bulbs",
    "Plastic Film",
    "Threads",
]


class AddressType(TypedDict):
    """Define a type to represent an address."""

    street1: str
    city: str
    state: str
    postal_code: str


@dataclass(frozen=True)
class RidwellAccount:  # pylint: disable=too-many-instance-attributes
    """Define a Ridwell account."""

    _async_request: Callable = field(compare=False)

    account_id: str
    address: AddressType
    email: str
    full_name: str
    phone: str
    subscription_id: str
    subscription_active: bool

    async def async_get_next_pickup_event(self) -> RidwellPickupEvent:
        """Get the next pickup event based on today's date."""
        pickup_events = await self.async_get_pickup_events()
        for event in pickup_events:
            if event.pickup_date >= date.today():
                return event
        raise RidwellError("No pickup events found after today")

    async def async_get_pickup_events(self) -> list[RidwellPickupEvent]:
        """Get pickup events for this subscription."""
        resp = await self._async_request(
            json={
                "operationName": "upcomingSubscriptionPickups",
                "variables": {"subscriptionId": self.subscription_id},
                "query": QUERY_SUBSCRIPTION_PICKUPS,
            },
        )

        return [
            RidwellPickupEvent(
                self._async_request,
                event_data["id"],
                datetime.strptime(event_data["pickupOn"], "%Y-%m-%d").date(),
                [
                    RidwellPickup(
                        titlecase(
                            pickup["pickupOfferPickupProduct"]["pickupOffer"][
                                "category"
                            ]["name"]
                        ),
                        pickup["pickupOfferPickupProduct"]["pickupOffer"]["id"],
                        pickup["pickupOfferPickupProduct"]["pickupOffer"]["priority"],
                        pickup["pickupOfferPickupProduct"]["pickupProduct"]["id"],
                        pickup["quantity"],
                    )
                    for pickup in event_data["pickupProductSelections"]
                ],
                event_data["state"],
            )
            for event_data in resp["data"]["upcomingSubscriptionPickups"]
        ]


@dataclass(frozen=True)
class RidwellPickup:
    """Define a Ridwell pickup (i.e., the thing being picked up)."""

    name: str
    offer_id: str
    priority: int
    product_id: str
    quantity: int

    category: Literal["add_on", "rotating", "standard"] = field(init=False)

    def __post_init__(self) -> None:
        """Perform some post-init init."""
        if self.name in PICKUP_TYPES_ADD_ON:
            category = CATEGORY_ADD_ON
        elif self.name in PICKUP_TYPES_STANDARD:
            category = CATEGORY_STANDARD
        else:
            category = CATEGORY_ROTATING
        object.__setattr__(self, "category", category)


@dataclass(frozen=True)
class RidwellPickupEvent:
    """Define a Ridwell pickup event."""

    _async_request: Callable = field(compare=False)
    _event_id: str

    pickup_date: date
    pickups: list[RidwellPickup]
    state: Literal["initialized", "scheduled"]

    async def async_get_estimated_cost(self) -> float:
        """Get the estimated cost (USD) of this pickup based on its pickup types."""
        if not self.pickups:
            return 0.0

        resp = await self._async_request(
            json={
                "operationName": "subscriptionPickupQuote",
                "variables": {
                    "input": {
                        "subscriptionPickupId": self._event_id,
                        "addOnSelections": [
                            {
                                "productId": pickup.product_id,
                                "offerId": pickup.offer_id,
                                "quantity": pickup.quantity,
                            }
                            for pickup in self.pickups
                        ],
                    }
                },
                "query": QUERY_SUBSCRIPTION_PICKUP_QUOTE,
            },
        )

        return cast(float, resp["data"]["subscriptionPickupQuote"]["totalCents"] / 100)


def decode_jwt(encoded_jwt: str) -> dict[str, Any]:
    """Decode and return a JWT."""
    return cast(
        Dict[str, Any],
        jwt.decode(
            encoded_jwt,
            "secret",
            algorithms=["HS256"],
            options={"verify_signature": False},
        ),
    )


class Client:  # pylint: disable=too-many-instance-attributes
    """Define the client."""

    def __init__(
        self,
        email: str,
        password: str,
        *,
        request_retries: int = DEFAULT_RETRIES,
        request_retry_delay: int = DEFAULT_RETRY_DELAY,
        session: ClientSession | None = None,
    ) -> None:
        """Initialize.

        Note that this is not intended to be instantiated directly; instead, users
        should use the async_get_client coroutine method.
        """
        self._email = email
        self._password = password
        self._request_retries = request_retries
        self._request_retry_delay = request_retry_delay
        self._session = session

        # Intended to be filled in after login:
        self._token: str | None = None
        self.user_id: str | None

    async def async_authenticate(self) -> None:
        """Authenticate the API."""
        resp = await self.async_request(
            json={
                "operationName": "createAuthentication",
                "variables": {
                    "input": {"emailOrPhone": self._email, "password": self._password}
                },
                "query": QUERY_AUTH_DATA,
            },
        )
        self._token = resp["data"]["createAuthentication"]["authenticationToken"]
        assert self._token
        token_data = decode_jwt(self._token)
        self.user_id = token_data["ridwell/userId"]

    async def async_get_accounts(self) -> dict[str, RidwellAccount]:
        """Get all accounts for this user."""
        resp = await self.async_request(
            json={
                "operationName": "user",
                "variables": {"id": self.user_id},
                "query": QUERY_ACCOUNT_DATA,
            }
        )
        user_data = resp["data"]["user"]

        return {
            account["id"]: RidwellAccount(
                self.async_request,
                account["id"],
                account["address"],
                user_data["email"],
                user_data["fullName"],
                user_data["phone"],
                account["activeSubscription"]["id"],
                account["activeSubscription"]["state"] == "active",
            )
            for account in user_data["accounts"]
        }

    async def async_request(self, **kwargs: dict[str, Any]) -> dict[str, Any]:
        """Make an API request (based on a Ridwell "query string")."""
        kwargs.setdefault("headers", {})

        use_running_session = self._session and not self._session.closed
        if use_running_session:
            session = self._session
        else:
            session = ClientSession(timeout=ClientTimeout(total=DEFAULT_TIMEOUT))

        assert session

        data: dict[str, Any] = {}
        retry = 0

        while retry < self._request_retries:
            if self._token:
                kwargs["headers"]["Authorization"] = f"Bearer {self._token}"
            async with session.request("post", API_BASE_URL, **kwargs) as resp:
                try:
                    data = await resp.json()
                    resp.raise_for_status()
                except (ClientError, ContentTypeError) as err:
                    raise RequestError(err) from err

                # Ridwell's API can return HTTP 200 responses that are still errors, so
                # we make sure to check for that:
                try:
                    raise_for_data_error(data)
                except TokenExpiredError:
                    retry += 1
                    LOGGER.info(
                        "Token failed; refreshing and trying again (attempt %s of %s)",
                        retry,
                        self._request_retries,
                    )
                    await self.async_authenticate()
                    await asyncio.sleep(self._request_retry_delay)
                    continue

                break
        else:
            # We only end up here if we continue to have credential issues after
            # several retries:
            raise InvalidCredentialsError("Unable to refresh access token") from None

        if not use_running_session:
            await session.close()

        LOGGER.debug(
            "Received data (operation: %s): %s",
            kwargs.get("json", {}).get("operationName"),
            data,
        )

        return data


async def async_get_client(
    email: str,
    password: str,
    *,
    request_retries: int = DEFAULT_RETRIES,
    request_retry_delay: int = DEFAULT_RETRY_DELAY,
    session: ClientSession | None = None,
) -> Client:
    """Get an authenticated client."""
    client = Client(
        email,
        password,
        request_retries=request_retries,
        request_retry_delay=request_retry_delay,
        session=session,
    )
    await client.async_authenticate()
    return client



"""Define package exceptions."""
from __future__ import annotations

from typing import Any


class RidwellError(Exception):
    """Define a base exception."""

    pass


class InvalidCredentialsError(RidwellError):
    """Define an error related to a bad credentials."""

    pass


class RequestError(RidwellError):
    """Define an error related to a bad HTTP request."""

    pass


class TokenExpiredError(RidwellError):
    """Define an error related to an expired token."""

    pass


DATA_ERROR_MAP = {
    "The password you entered is incorrect. Please try again.": InvalidCredentialsError,
    "login required": TokenExpiredError,
}


def raise_for_data_error(data: dict[str, Any]) -> None:
    """Raise an appropriate error if a message exists in the response data."""
    if "errors" not in data:
        return

    # A response payload may feasibly contain several errors (based on its JSON
    # structure), so we only handle the first one:
    first_error = data["errors"][0]
    message = first_error["message"]
    exc = DATA_ERROR_MAP.get(message, RequestError)
    raise exc(message)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['aioridwell']

package_data = \
{'': ['*']}

install_requires = \
['PyJWT==2.1.0',
 'aiohttp>=3.7.4,<4.0.0',
 'pytz>=2021.3,<2022.0',
 'titlecase>=2.3,<3.0']

setup_kwargs = {
    'name': 'aioridwell',
    'version': '0.2.0',
    'description': 'A Python3, asyncio-based API for interacting with Ridwell waste recycling',
    'long_description': '# ♻️ aioridwell: A Python3, asyncio-based API for interacting with Ridwell\n\n[![CI](https://github.com/bachya/aioridwell/workflows/CI/badge.svg)](https://github.com/bachya/aioridwell/actions)\n[![PyPi](https://img.shields.io/pypi/v/aioridwell.svg)](https://pypi.python.org/pypi/aioridwell)\n[![Version](https://img.shields.io/pypi/pyversions/aioridwell.svg)](https://pypi.python.org/pypi/aioridwell)\n[![License](https://img.shields.io/pypi/l/aioridwell.svg)](https://github.com/bachya/aioridwell/blob/master/LICENSE)\n[![Code Coverage](https://codecov.io/gh/bachya/aioridwell/branch/master/graph/badge.svg)](https://codecov.io/gh/bachya/aioridwell)\n[![Maintainability](https://api.codeclimate.com/v1/badges/9c1dcc1c991cecb06eda/maintainability)](https://codeclimate.com/github/bachya/aioridwell/maintainability)\n[![Say Thanks](https://img.shields.io/badge/SayThanks-!-1EAEDB.svg)](https://saythanks.io/to/bachya)\n\n`aioridwell` is a Python 3, asyncio-friendly library for interacting with\n[Ridwell](https://ridwell.com) to view information on upcoming recycling pickups.\n\n- [Installation](#installation)\n- [Python Versions](#python-versions)\n- [Usage](#usage)\n- [Contributing](#contributing)\n\n# Installation\n\n```python\npip install aioridwell\n```\n\n# Python Versions\n\n`aioridwell` is currently supported on:\n\n* Python 3.8\n* Python 3.9 \n\n# Usage\n\n## Creating and Using a Client\n\nThe `Client` is the primary method of interacting with the API:\n\n```python\nimport asyncio\n\nfrom aioridwell import async_get_client\n\n\nasync def main() -> None:\n    client = await async_get_client("<EMAIL>", "<PASSWORD>")\n    # ...\n\n\nasyncio.run(main())\n```\n\nBy default, the library creates a new connection to the API with each coroutine. If\nyou are calling a large number of coroutines (or merely want to squeeze out every second of runtime savings possible), an\n[`aiohttp`](https://github.com/aio-libs/aiohttp) `ClientSession` can be used for connection\npooling:\n\n```python\nimport asyncio\n\nfrom aiohttp import ClientSession\n\nfrom aiowatttime import Client\n\n\nasync def main() -> None:\n    async with ClientSession() as session:\n        client = await async_get_client("<EMAIL>", "<PASSWORD>", session=session)\n        # ...\n\n\nasyncio.run(main())\n```\n\n## Getting Accounts\n\nGetting all accounts associated with this email address is easy:\n\n```python\nimport asyncio\n\nfrom aioridwell import async_get_client\n\n\nasync def main() -> None:\n    client = await async_get_client("<EMAIL>", "<PASSWORD>")\n\n    accounts = await client.async_get_accounts()\n    # >>> {"account_id_1": RidwellAccount(...), ...}\n\n\nasyncio.run(main())\n```\n\nThe `RidwellAccount` object comes with some useful properties:\n\n* `account_id`: the Ridwell ID for the account\n* `address`: the address being serviced\n* `email`: the email address on the account\n* `full_name`: the full name of the account owner\n* `phone`: the phone number of the account owner\n* `subscription_id`: the Ridwell ID for the primary subscription\n* `subscription_active`: whether the primary subscription is active\n\n## Getting Pickup Events\n\nGetting pickup events associated with an account is easy, too:\n\n```python\nimport asyncio\n\nfrom aioridwell import async_get_client\n\n\nasync def main() -> None:\n    client = await async_get_client("<EMAIL>", "<PASSWORD>")\n\n    accounts = await client.async_get_accounts()\n    for account in accounts.values():\n        events = await account.async_get_pickup_events()\n        # >>> [RidwellPickupEvent(...), ...]\n\n        # You can also get just the next pickup event from today\'s date:\n        next_event = await account.async_get_next_pickup_event()\n        # >>> RidwellPickupEvent(...)\n\n\nasyncio.run(main())\n```\n\nThe `RidwellPickupEvent` object comes with some useful properties:\n\n* `pickup_date`: the date of the pickup (in `datetime.date` format)\n* `pickups`: a list of `RidwellPickup` objects\n* `state`: either `initialized` (not scheduled for pickup) or `scheduled`\n\nLikewise, the `RidwellPickup` object comes with some useful properties:\n\n* `category`: the category of the pickup (`standard`, `rotating`, or `add_on`)\n* `name`: the name of the item being picked up\n* `offer_id`: the Ridwell ID for this particular offer\n* `priority`: the pickup priority\n* `product_id`: the Ridwell ID for this particular product\n* `quantity`: the amount of the product being picked up\n\n### Calculating a Pickup Event\'s Esimated Cost\n\nCalculating the estimated cost of a pickup event is, you guessed it, easy:\n\n```python\nimport asyncio\n\nfrom aioridwell import async_get_client\n\n\nasync def main() -> None:\n    client = await async_get_client("<EMAIL>", "<PASSWORD>")\n\n    accounts = await client.async_get_accounts()\n    for account in accounts.values():\n        events = await account.async_get_pickup_events()\n        event_1_cost = await events[0].async_get_estimated_cost()\n        # >>> 22.00\n\n\nasyncio.run(main())\n```\n\n# Contributing\n\n1. [Check for open features/bugs](https://github.com/bachya/aioridwell/issues)\n  or [initiate a discussion on one](https://github.com/bachya/aioridwell/issues/new).\n2. [Fork the repository](https://github.com/bachya/aioridwell/fork).\n3. (_optional, but highly recommended_) Create a virtual environment: `python3 -m venv .venv`\n4. (_optional, but highly recommended_) Enter the virtual environment: `source ./.venv/bin/activate`\n5. Install the dev environment: `script/setup`\n6. Code your new feature or bug fix.\n7. Write tests that cover your new functionality.\n8. Run tests and ensure 100% code coverage: `script/test`\n9. Update `README.md` with any new documentation.\n10. Add yourself to `AUTHORS.md`.\n11. Submit a pull request!\n',
    'author': 'Aaron Bach',
    'author_email': 'bachya1208@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/bachya/aioridwell',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8.0,<4.0.0',
}


setup(**setup_kwargs)
